! Copyright (c) 2017, Tsutomu Kato, SofTek Systems. Inc. All rights reserved.
!
! Permission is hereby granted, free of charge, to any person obtaining a
! copy of this software and associated documentation files (the "Software"),
! to deal in the Software without restriction, including without limitation
! the rights to use, copy, modify, merge, publish, distribute, sublicense,
! and/or sell copies of the Software, and to permit persons to whom the
! Software is furnished to do so, subject to the following conditions:
!
! The above copyright notice and this permission notice shall be included in
! all copies or substantial portions of the Software.
!
! THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
! IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
! FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL
! THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
! LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
! FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER
! DEALINGS IN THE SOFTWARE.
!

module SP_solvers
    use precision
    implicit none

  contains

    real(fp_kind) function vec_max(n, x)
    implicit none
    integer       :: i, n
    real(fp_kind) :: x(n)
      vec_max = 0.0_fp_kind 
      do i = 1, n
      vec_max = max( abs(x(i)), vec_max )
      end do
    end function

    real(fp_kind) function csr_mat_max(m, n, nz, csrValA, csrRowPtrA, csrColIndA)
    implicit none
    integer       :: m, n, nz
    integer       :: csrRowPtrA(m+1), csrColIndA(nz)
    real(fp_kind) :: csrValA(nz)
    real(fp_kind) :: sum, norminf
    integer       :: i, j, start, end
    norminf = 0.0_fp_kind

    do i = 1, m
      sum = 0.0_fp_kind
      start = csrRowPtrA(i  )
      end   = csrRowPtrA(i+1) - 1
        do j = start, end
          sum = sum + abs(csrValA(j))
        end do
      norminf = max (norminf, sum)
    end do
    csr_mat_max = norminf
    end function
         
  
     

!=============================================================
    subroutine cusolverSPs(gpuonly, solver, reorder, m, n, nz, csrRowPtrA, csrColIndA, csrValA, x, rhs, &
                          tol, index_base)
!-------------------------------------------------------------
!   Bi-Conjugate Gradient Stabilized (BiCGStab)
!   A is an n × n sparse matrix defined in CSR storage format
!-------------------------------------------------------------
    use measure_time
    use cusparse
    use cuSOLVER_SP_Solver
    implicit none

! Arrays for CUDA cuSPARSE CSR format
    integer       :: m                ! number of rows of matrix
    integer       :: n                ! number of columns of matrix
    integer       :: nz               ! nonzero elements of A
    integer       :: csrRowPtrA(n+1)  ! the array of offsets corresponding to the start of each row
    integer       :: csrColIndA (nz)  ! sorted by row and by column within each row
    real(fp_kind) :: csrValA    (nz)  ! sorted by row and by column within each row
                                      ! B = Q*A*Q^T
    real(fp_kind) :: x  (n)           ! initial guess & result
    real(fp_kind) :: rhs(n)           ! right hand side
    real(fp_kind) :: b  (n)           ! a copy of rhs
    character(c_char), allocatable :: buffer(:) 
    integer       :: index_base

    integer       :: gpuonly          ! 0 = both CPU and GPU, 1 = GPU only
    integer       :: solver           ! 1 = cholesky, 2 = LU, 3 = QR
    integer       :: reorder          ! manula reordering 1 = symrcm, 2 = symamd reordering
    integer       :: singularity      ! -1 if A is invertible under tol
    integer       :: issym
    real(fp_kind) :: rsum, diff, error
    real(fp_kind) :: x_inf, r_inf, A_inf

    real(fp_kind) :: tol              !  to decide if singular or not

! Local variables
    type(cusolverSpHandle)   :: cusolver_H
    type(cusparseHandle)     :: cusparse_H
    type(cusparseMatDescr)   :: descrA
    integer                  :: cusolver_status
    integer                  :: status, ierr_code
    real(fp_kind),parameter::  one = 1.0, zero = 0.0, onem1 = -1.0

    integer                  :: i, j, k
    integer                  :: nprint
! Parameters
    integer, parameter       :: idbg = 0
    real(fp_kind),parameter  :: eps = 6.9d-310     ! machine epsilon

! Initialize
    singularity = 0
    issym   = 0

    if (m /= n) then
      print *,"Error: only support square matrix."
      stop
    end if 
! 
    print '(/1x,a)',"=================================================================="
    if (solver == 1) then
      print '(1x,a)', "The linear system is solved by sparse CHOLESKY factorization."
    else if (solver == 2) then
      print '(1x,a)', "The linear system is solved by sparse LU factorization."
    else if (solver == 3) then
      print '(1x,a)', "The linear system is solved by sparse QR factorization."
    end if
    if (reorder == 0) then
      print '(1x,a)', "Reordering scheme : No reordering"
    else if (reorder == 1) then
      print '(1x,a)', "Reordering scheme : symrcm"
    else if (reorder == 2) then
      print '(1x,a)', "Reordering scheme : symamd"
    end if
    print '(/1x,a,d15.10,2x,a,i8/)',"Tolerance to decide if singular or not = ", tol

!   set b = rhs, where Ax = rhs (Ax = b)
    b(:) = rhs(:)

    k = 0

! Create cusolver/cusparse handle, and matrix descriptor

    status = cusolverSpCreate(cusolver_H)
    status = cusparseCreate(cusparse_H)
    status = cusparseCreateMatDescr(descrA)
    status = cusparseSetMatType(descrA, CUSPARSE_MATRIX_TYPE_GENERAL)
    status = cusparseSetMatDiagType(descrA, CUSPARSE_DIAG_TYPE_NON_UNIT)
    status = cusparseSetMatIndexBase(descrA, CUSPARSE_INDEX_BASE_ONE) ! base=1

! Verify if A has symmetric pattern or not
    status = cusolverSpXcsrissymHost( &
        cusolver_H, m, nz, descrA, csrRowPtrA(1), csrRowPtrA(2), csrColIndA, issym)
    if (issym == 0) then
      print *, "A matrix has no symmmetric pattern."
    else
      print *, "A matrix has symmmetric pattern."
    end if 
    print '(1x,a)',"=================================================================="
    if(idbg .eq. 1) print *, "cusolverSpXcsrissymHost-status=", status

! Clock init
    status = init_clock
    if (gpuonly == 1 ) goto 100

! Clock start
       status = start_clock

    if ( solver == 1 ) then                                                 ! cholesky
      status = cusolverSpDcsrlsvcholHost(cusolver_H, m, nz, &
               descrA, csrValA, csrRowPtrA, csrColIndA, &
               b, tol, reorder, x, singularity)
      if(idbg .ge. 1)  then
        print *
        print *,"singularity = ", singularity
        print *,"cusolverSpDcsrlsvcholHost-status = ", status
      end if

    else if ( solver == 2 ) then                                            ! LU
      status = cusolverSpDcsrlsvluHost(cusolver_H, m, nz, &
               descrA, csrValA, csrRowPtrA, csrColIndA, &
               b, tol, reorder, x, singularity)
      if(idbg .ge. 1)  then
        print *
        print *,"cusolverSpDcsrlsvluHost-status = ", status
        print *,"singularity = ", singularity
      end if

    else if ( solver == 3 ) then                                            ! QR
      status = cusolverSpDcsrlsvqrHost(cusolver_H, m, nz, &
               descrA, csrValA, csrRowPtrA, csrColIndA, &
               b, tol, reorder, x, singularity)
      if(idbg .eq. 1) then
        print *
        print *,"cusolverSpDcsrlsvqrHost-status = ", status
        print *,"singularity = ", singularity
      end if
    else
       print *,"Error: Solver",solver," is unknown solver"
    end if

       status = end_clock
       time_tag = "cuSOLVER SP solver result on [ CPU ]"
       status = print_time(time_tag)

!$acc data copy(b, X)  copyin( csrValA, csrRowPtrA, csrColIndA ) 
  !$acc host_data use_device(b, x, csrValA, csrRowPtrA, csrColIndA)
    ! evaluate residual b = b - A*x (result on CPU)
    status = cusparseDcsrmv(cusparse_H, CUSPARSE_OPERATION_NON_TRANSPOSE, &
             m, n, nz, onem1, descrA, csrValA, csrRowPtrA, csrColIndA, x, one, b)
  !$acc end host_data
!$acc end data

    if ( singularity .ge. 0 ) &
    print '(/2x,a,i5,,2x,a, d10.3/)', "WARNING!! : the matrix is singular at row", singularity+1, "under tolerance", tol
    r_inf = vec_max (m, b)
    x_inf = vec_max (n, x)
    A_inf =csr_mat_max(m, n, nz, csrValA, csrRowPtrA, csrColIndA)
    print '(2x,a,d20.13)',"|b - A*x| =", r_inf
    print '(2x,a,d20.13)',"      |A| =", A_inf
    print '(2x,a,d20.13)',"      |x| =", x_inf
    print '(2x,a,d20.13)',"|b - A*x|/(|A|*|x|) =", r_inf/(A_inf * x_inf)

! output result
    print *
    nprint = n
    if ( n > 10 ) nprint = 10
    do i = 1, nprint
       print '(1x, a,i7,a,e18.12)', "x(", i, ")=", X(i)
    end do
    print *,""
    do i = n-nprint, n
       print '(1x, a,i7,a,e18.12)', "x(", i, ")=", X(i)
    end do

!   reset b = rhs, where Ax = rhs (Ax = b)
    b(:) = rhs(:)
    x = 0.0_fp_kind

100 continue

! Clock start
      status = start_clock

!$acc data copy(b, X)  copyin( csrValA, csrRowPtrA, csrColIndA ) 
!$acc host_data use_device(b, x, csrValA, csrRowPtrA, csrColIndA)

    if ( solver == 1 ) then                                                 ! cholesky
      status = cusolverSpDcsrlsvchol(cusolver_H, m, nz, &
               descrA, csrValA, csrRowPtrA, csrColIndA, &
               b, tol, reorder, x, singularity)
      if(idbg .eq. 1)  then
        print *
        print *,"singularity = ", singularity
        print *,"cusolverSpDcsrlsvchol-status = ", status
      end if

    else if ( solver == 2 ) then                                            ! LU
      print '(/a)'," LU SP solver for GPU is not available now."
      print *,"-------------------------------------------"
      stop

    else if ( solver == 3 ) then                                            ! QR
      status = cusolverSpDcsrlsvqr(cusolver_H, m, nz, &
               descrA, csrValA, csrRowPtrA, csrColIndA, &
               b, tol, reorder, x, singularity)
      if(idbg .eq. 1) then
        print *
        print *,"cusolverSpDcsrlsvqr-status = ", status
        print *,"singularity = ", singularity
      end if
    else
       print *,"Error: Solver",solver," is unknown solver"
    end if

!$acc end host_data
!$acc end data

      status = end_clock
      time_tag = "cuSOLVER SP solver result on [ GPU ]"
      status = print_time(time_tag)

!$acc data copy(b, X)  copyin( csrValA, csrRowPtrA, csrColIndA ) 
  !$acc host_data use_device(b, x, csrValA, csrRowPtrA, csrColIndA)
    ! evaluate residual b = b - A*x (result on CPU)
    status = cusparseDcsrmv(cusparse_H, CUSPARSE_OPERATION_NON_TRANSPOSE, &
             m, n, nz, onem1, descrA, csrValA, csrRowPtrA, csrColIndA, x, one, b)
  !$acc end host_data
!$acc end data

    if ( singularity .ge. 0 ) &
      print '(/2x,a,i5,,2x,a, d10.3/)', "Warning!! : the matrix is singular at row", singularity+1, "under tolerance", tol
    r_inf = vec_max (m, b)
    x_inf = vec_max (n, x)
    A_inf =csr_mat_max(m, n, nz, csrValA, csrRowPtrA, csrColIndA)
    print '(2x,a,d20.13)',"|b - A*x| =", r_inf
    print '(2x,a,d20.13)',"      |A| =", A_inf
    print '(2x,a,d20.13)',"      |x| =", x_inf
    print '(2x,a,d20.13)',"|b - A*x|/(|A|*|x|) =", r_inf/(A_inf * x_inf)

! output result 
    print *
    nprint = n
    if ( n > 10 ) nprint = 10
    do i = 1, nprint
       print '(1x, a,i7,a,e18.12)', "x(", i, ")=", X(i)
    end do
    print *,""
    do i = n-nprint, n
       print '(1x, a,i7,a,e18.12)', "x(", i, ")=", X(i)
    end do

    status = cusparseDestroy(cusparse_H)
    status = cusparseDestroyMatDescr(descrA)

    end subroutine cusolverSPs

End module SP_solvers

